#     Copyright (c) 2013-2016 OSIsoft, LLC. All rights reserved.
# 
#     THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
#     OSIsoft, LLC.  USE, DISCLOSURE, OR REPRODUCTION IS PROHIBITED WITHOUT
#     THE PRIOR EXPRESS WRITTEN PERMISSION OF OSIsoft, LLC.
# 
#     RESTRICTED RIGHTS LEGEND
#     Use, duplication, or disclosure by the Government is subject to restrictions
#     as set forth in subparagraph (c)(1)(ii) of the Rights in Technical Data and
#     Computer Software clause at DFARS 252.227.7013
# 
#     OSIsoft, LLC
#     777 Davis Street, Suite 250, San Leandro CA 94577

require 'fluent/input'
require 'rest-client'
require 'json'
require 'time'

module Fluent
  class Piin < Input
    # First, register the plugin. NAME is the name of this plugin
    # and identifies the plugin in the configuration file.
    Fluent::Plugin.register_input('piin', self)

    stop_flag = false
    webid = nil
    piwebapiurl = 'localhost', tagpath = nil 
    tag = nil
    regetrate = 10

    # This method is called before starting.
    # 'conf' is a Hash that includes configuration parameters.
    # If the configuration is invalid, raise Fluent::ConfigError.
    def configure(conf)
      super
      @piwebapiurl = conf['piwebapiurl']
      (conf.key?('tagpath')) ? @tagpath = conf['tagpath'] : @tagpath = nil
      (conf.key?('webid')) ? @webid = conf['webid'] : @webid = nil
      (conf.key?('username')) ? @username = conf['username'] : @username = nil    
      (conf.key?('pwd')) ? @pwd = conf['pwd'] : @pwd = nil
      (conf.key?('anon')) ? @anon = conf['anon'] : @anon = false
      (conf.key?('tag')) ? @tag = conf['tag'] : @tag = 'piin'
      (conf.key?('regetrate')) ? @regetrate = conf['regetrate'] : @regetrate = 10
    end

    # This method is called when starting.
    # Open sockets or files and create a thread here.
    def start
      super
      
      if(@webid == nil)
        url = "https://#{@piwebapiurl}/piwebapi/points/?path=#{@tagpath}"
        begin
          client = RestClient::Resource.new(url, :verify_ssl => OpenSSL::SSL::VERIFY_NONE, :user => @username, :password =>@pwd)
          response = client.get
          @webid = (JSON.parse(response.body))['WebId']
        rescue
          log.warn "Errors. Web Id not found. Not able to bring in data from the specified address"
          log.warn url
          @webid = nil
        end
      end
      
      @stop_flag = false
      @thread = Thread.new(&method(:thread_main))
    end

    # This method is called when shutting down.
    # Shutdown the thread and close sockets or files here.
    def shutdown    
      super  
      @stop_flag = true
      $log.debug "Waiting for thread to finish"
      @thread.join
    end   

    def thread_main
      if(@webid == nil)
        return
      end
      url = "https://#{@piwebapiurl}/piwebapi/streams/#{@webid}/value"
      client = RestClient::Resource.new(url, :verify_ssl => OpenSSL::SSL::VERIFY_NONE,:user => @username, :password =>@pwd)
      until @stop_flag
        sleep @regetrate
        begin
            response = client.get :content_type => 'application/json'
            responseObj = JSON.parse(response)
            time = Time.parse(responseObj['Timestamp'])
            if (time == nil)
              time = Engine.now
            end
            router.emit(@tag, time, responseObj)
        rescue => e
            log.warn "Having problems connecting to PI WebAPI.  Will try again."
        end
      end      
    end
    
  end
end